//==================================================================================================
/*
  EVE - Expressive Vector Engine
  Copyright : EVE Project Contributors
  SPDX-License-Identifier: BSL-1.0
*/
//==================================================================================================

//
// In this example we will have a look at the same problem as in writing_new/collect_indexes__complicated_real_example
// but we will write a loop instead of using a zip with iota because it's usefult to understand.
//

#include <eve/algo/concepts.hpp>
#include <eve/algo/preprocess_range.hpp>

#include <eve/module/core.hpp>

#include <concepts>
#include <type_traits>
#include <vector>

template <
  eve::algo::relaxed_range R, // relaxed_range is our version of the std::range
                              // concept that supports things like aligned pointers, zip, etc.
                              // ---
  typename P,                 // Predicate. Don't conceptify, since it's quite difficult (see below)
                              // and does not seem that useful.
                              // ---
  std::integral IdxType,      // Index type, should be big enough to store any element's position
                              // ---
  typename Alloc              // To not have to deal with allocators ourselves,
                              // We'll just allow you to pass a vector with any allocator.
>
void collect_indexes(R&& r, P p, std::vector<IdxType, Alloc>& res)
{
  EVE_ASSERT((static_cast<std::size_t>(r.end() - r.begin()) <= std::numeric_limits<IdxType>::max()),
              "The output type is not big enough to hold potential indexes");

  // Prepare the output in case it was not empty.
  res.clear();

  // We have to do this everywhere in `algo`.
  // There is a requirement for `eve::load` that the pointer/iterator is to
  // somewheere in valid memory, even if we ignore everything.
  if (r.begin() == r.end()) return;

  // This converts the input to eve's understanding of a range:
  // unwraps vector::iterator to pointers, things like that.
  // This is also where all of our dealing with traits happen,
  // passing `consider_types<IdxType>` will make sure that we choose the cardinal (width)
  // appropriate for both range type and index type.
  auto processed = eve::algo::preprocess_range(
    eve::algo::traits{eve::algo::consider_types<IdxType>}, std::forward<R>(r)
  );

  // Here we would normally use `for_each_iteration` but it's good to understand what's
  // happening inside.
  auto  f  = processed.begin();
  auto  l  = processed.end();
  using I  = decltype(f);
  using N  = eve::iterator_cardinal_t<I>;
  // ^ Because this is the first place where we get the cardinal,
  //   we couldn't specify the requirement on the predicate

  auto precise_l = f + ((l - f) / N{}()) * N{}();

  // We are going to overallocate the output so that we can don't
  // have to store partial registers.
  res.resize(l - f + N{}());
  IdxType* out = res.data();

  // In a normal loop this would be i from for (i = 0; i != ...)
  // The lambda constructor will fill in (0, 1, 2, ...);
  eve::wide<IdxType, N> wide_i{[](int i, int) { return i; }};

  while (f != precise_l) {
    auto test = p(eve::load(f));  // apply the predicate

    // Compress store is the working horse of `remove`. It get's values, mask and where to store.
    // Writes all of the elements, for which mask is true.
    // Returns a pointer to after the last stored.
    // `unsafe` refers to the fact that it's allowed to store up to the whole register,
    // as long as the first elements are fine.
    out = eve::unsafe(eve::compress_store)(wide_i, test, out);

    wide_i += N{}();  // ++i
    f      += N{}();  // ++f
  }

  // Deal with tail
  if (f != l) {
    auto ignore = eve::keep_first(l - f);  // elements after l should not be touched
    auto test   = p(eve::load[ignore](f)); // This will safely load the partial register.
                                           // The last elements that correspond to after l will be garbage.

    // We have overallocated the output, but we still need to mask out garbage elements
    test = test && ignore.mask(eve::as(test));
    out  = eve::unsafe(eve::compress_store)(wide_i, test, out);
  }

  // Clean up the vector
  res.resize(out - res.data());
  res.shrink_to_fit();
}

// ---------------------------------------------------

#include "test.hpp"

#include <eve/algo/container/soa_vector.hpp>
#include <eve/memory/aligned_allocator.hpp>

#include "unit/algo/algo_test.hpp"

TTS_CASE("collect_indexes, elements equal to 2")
{
  std::vector<int>      input    = { 1, 1, 2, -3, 2, 10};
  std::vector<unsigned> expected = { 2, 4 };
  std::vector<unsigned> actual;
  collect_indexes(input, [](auto x) { return x == 2; }, actual);
  TTS_EQUAL(expected, actual);
};

TTS_CASE("collect_indexes for objects")
{
  using obj           = kumi::tuple<int, double>;

  eve::algo::soa_vector<obj> objects {obj{1, 2.0}, obj{-1, 2.0}, obj{1, 1}};

  std::vector<std::uint16_t> expected{0, 2};
  std::vector<std::uint16_t> idxs;
  collect_indexes(objects, [](auto x) { return get<0>(x) > 0; }, idxs);

  TTS_EQUAL(idxs, expected);
};

TTS_CASE("collect_indexes clears the result")
{
  std::vector<int>      input;
  std::vector<unsigned> expected;
  std::vector<unsigned> actual(65u);
  collect_indexes(input, [](auto x) { return x == 2; }, actual);
  TTS_EQUAL(expected, actual);
};

// ---------------
// push data through test

struct collect_indexes_generic_test
{
  std::vector<std::int64_t> expected;

  void init(auto*, auto* f, auto* l, auto*)
  {
    int count = 0;

    while (f != l) {
      if (count % 3 == 0) *f = 1;
      else                *f = 0;
      ++f;
    }
  }

  void run(auto rng)
  {
    expected.clear();
    for (auto* f = eve::unalign(rng.begin()); f != rng.end(); ++f) {
      if (*f == 1) expected.push_back(f - rng.begin());
    }

    std::vector<std::int64_t> actual;
    collect_indexes(rng, [](auto x) { return x == 1; }, actual);
    TTS_EQUAL(expected, actual);
  }

  void adjust(auto*, auto* f, auto* l, auto* page_end) const
  {
    *f = 1;
    if (l != page_end) *l = 1;
  }
};

TTS_CASE_TPL("Check collect indexes, lots", eve::test::scalar::all_types)
<typename T>(tts::type<T>)
{
  using tgt_t = eve::wide<T, eve::fixed<eve::expected_cardinal_v<std::int64_t>>>;
  algo_test::page_ends_test(eve::as<tgt_t>{}, collect_indexes_generic_test{});
};
